

from .warp import attach_spm_warp_fmri_wf
from .clean import attach_fmri_cleanup_wf
from .rest import attach_rest_preprocessing, attach_rest_grptemplate_preprocessing
from .grouptemplate import attach_spm_fmri_grouptemplate_wf